using System;
using System.Globalization;
using UnityEngine;

namespace DoesNotCat {
    public class RealFakeLabias : MVRScript {

        private const string pluginName = "RealFakeLabias";
        private readonly string[] morphNames = {
            "Labia majora-spread-LLow",
            "Labia majora-spread-LMid",
            "Labia majora-spread-LUp",
            "Labia majora-spread-RLow",
            "Labia majora-spread-RMid",
            "Labia majora-spread-RUp",
            "Labia minora-spread-LLow",
            "Labia minora-spread-LMid",
            "Labia minora-spread-LUp",
            "Labia minora-spread-RLow",
            "Labia minora-spread-RMid",
            "Labia minora-spread-RUp"
        };

        private Atom atom;
        private DAZMorph[] morphs;
        private DAZBone lThigh, rThigh, pelvis;
        private int nMorphs;
        public JSONStorableFloat minSpread, maxSpread, spreading, spreadingOuter, spreadingInner, spreadingFront, spreadingMid, spreadingBack, bias, biasOuter, biasInner, biasFront, biasMid, biasBack;
        public JSONStorableFloat[] morphValues;

        public override void Init () {
            atom = GetContainingAtom ();
            if (atom.type != "Person") {
                LogError ($"Plugin only works on atoms with type 'Person', but found an atom of type '{containingAtom.type}'. Plugin will not work on this atom.");
                enabledJSON.SetVal (false);
                return;
            }
            JSONStorable geo = atom.GetStorableByID ("geometry");
            DAZCharacterSelector character = geo as DAZCharacterSelector;
            DAZBones bones = character.rootBones;
            lThigh = bones.GetDAZBone ("lThigh");
            rThigh = bones.GetDAZBone ("rThigh");
            pelvis = bones.GetDAZBone ("pelvis");
            if (lThigh == null || rThigh == null || pelvis == null) {
                LogError ($"One of the necessary bones 'lThight', 'rThight' or 'pelvis' could not be found on this atom. Plugin will not work on this atom.");
                enabledJSON.SetVal (false);
                return;
            }

            // Obtain morphs
            nMorphs = morphNames.Length;
            morphs = new DAZMorph[nMorphs];
            GenerateDAZMorphsControlUI mc = character.morphsControlUI;
            for (int i = 0; i < nMorphs; ++i) {
                morphs[i] = mc.GetMorphByDisplayName (morphNames[i]);
                if (morphs[i] == null) {
                    LogError ($"The necessary morph '{morphNames[i]}' could not be found on this atom. Plugin will not work on this atom.");
                    enabledJSON.SetVal (false);
                    return;
                }
                if (morphs[i].isDriven && morphs[i].drivenBy != pluginName) {
                    LogError ($"The necessary morph '{morphNames[i]}' is already driven by '{morphs[i].drivenBy}'. Plugin will not work on this atom.");
                    enabledJSON.SetVal (false);
                    return;
                }
            }

            // Parameters + UI
            minSpread = CreateFloatParam ("Minimum morph change", -0.3f, -2f, 0f, false, val => { if (maxSpread.val < val) maxSpread.val = val; });
            maxSpread = CreateFloatParam ("Maximum morph change", 1, 0, 2, true, val => { if (minSpread.val > val) minSpread.val = val; });
            spreading = CreateFloatParam ("Spreading factor", 1f, 0f, 2f);
            spreadingOuter = CreateFloatParam ("Spreading factor outer labia", 0.7f, 0f, 2f);
            spreadingInner = CreateFloatParam ("Spreading factor inner labia", 0.3f, 0f, 2f);
            spreadingFront = CreateFloatParam ("Spreading factor front", 0.8f, 0f, 2f, true);
            spreadingMid = CreateFloatParam ("Spreading factor mid", 1f, 0f, 2f, true);
            spreadingBack = CreateFloatParam ("Spreading factor back", 0.8f, 0f, 2f, true);

            CreateSpacer (false);
            CreateSpacer (true);

            bias = CreateFloatParam ("Bias", 0f, -1f, 1f);
            biasOuter = CreateFloatParam ("Bias outer labia", 0f, -1f, 1f);
            biasInner = CreateFloatParam ("Bias inner labia", 0f, -1f, 1f);
            biasFront = CreateFloatParam ("Bias front", 0f, -1f, 1f, true);
            biasMid = CreateFloatParam ("Bias mid", 0f, -1f, 1f, true);
            biasBack = CreateFloatParam ("Bias back", 0f, -1f, 1f, true);

            CreateSpacer (false);
            CreateSpacer (true);

            // Store morph values
            morphValues = new JSONStorableFloat[nMorphs];
            for (int i = 0; i < nMorphs; ++i) {
                morphValues[i] = new JSONStorableFloat ($"{morphNames[i]} offset", 0f, -1f, 1f, false, true);
                morphValues[i].SetVal (morphs[i].morphValue);
                RegisterFloat (morphValues[i]);
                CreateSlider (morphValues[i], i >= nMorphs / 2);
            }

            UIDynamicButton resetButton = CreateButton ("Reset offsets to zero");
            resetButton.button.onClick.AddListener (() => {
                for (int i = 0; i < nMorphs; ++i) {
                    morphValues[i].SetValToDefault ();
                }
            });
        }

        public void Start () {
            // Lock morphs
            foreach (DAZMorph morph in morphs) {
                morph.SetDriven (true, pluginName, true);
            }
        }

        private void Update () {
            Vector3 left = lThigh.transform.up;
            Vector3 right = rThigh.transform.up;
            Vector3 forward = pelvis.transform.forward;
            Vector3 sideward = pelvis.transform.right;
            // Scalar projection
            float lForward = Vector3.Dot (left, forward);
            float lSideward = Vector3.Dot (left, sideward);
            float rForward = Vector3.Dot (right, forward);
            float rSideward = -Vector3.Dot (right, sideward);
            float lLow = Math.Abs (-0.5f - lForward) * lSideward * spreadingBack.val + bias.val + biasBack.val;
            float lMid = lSideward * spreadingMid.val + bias.val + biasMid.val;
            float lUp = Math.Abs (1f - lForward) * lSideward * spreadingFront.val + bias.val + biasFront.val;
            float rLow = Math.Abs (-0.5f - rForward) * rSideward * spreadingBack.val + bias.val + biasBack.val;
            float rMid = rSideward * spreadingMid.val + bias.val + biasMid.val;
            float rUp = Math.Abs (1f - rForward) * rSideward * spreadingFront.val + bias.val + biasFront.val;
            morphs[0].SetValue (clamp (lLow * spreading.val * spreadingOuter.val + biasOuter.val, minSpread.val, maxSpread.val) + morphValues[0].val);
            morphs[1].SetValue (clamp (lMid * spreading.val * spreadingOuter.val + biasOuter.val, minSpread.val, maxSpread.val) + morphValues[1].val);
            morphs[2].SetValue (clamp (lUp * spreading.val * spreadingOuter.val + biasOuter.val, minSpread.val, maxSpread.val) + morphValues[2].val);
            morphs[3].SetValue (clamp (rLow * spreading.val * spreadingOuter.val + biasOuter.val, minSpread.val, maxSpread.val) + morphValues[3].val);
            morphs[4].SetValue (clamp (rMid * spreading.val * spreadingOuter.val + biasOuter.val, minSpread.val, maxSpread.val) + morphValues[4].val);
            morphs[5].SetValue (clamp (rUp * spreading.val * spreadingOuter.val + biasOuter.val, minSpread.val, maxSpread.val) + morphValues[5].val);
            morphs[6].SetValue (clamp (lLow * spreading.val * spreadingInner.val + biasInner.val, minSpread.val, maxSpread.val) + morphValues[6].val);
            morphs[7].SetValue (clamp (lMid * spreading.val * spreadingInner.val + biasInner.val, minSpread.val, maxSpread.val) + morphValues[7].val);
            morphs[8].SetValue (clamp (lUp * spreading.val * spreadingInner.val + biasInner.val, minSpread.val, maxSpread.val) + morphValues[8].val);
            morphs[9].SetValue (clamp (rLow * spreading.val * spreadingInner.val + biasInner.val, minSpread.val, maxSpread.val) + morphValues[9].val);
            morphs[10].SetValue (clamp (rMid * spreading.val * spreadingInner.val + biasInner.val, minSpread.val, maxSpread.val) + morphValues[10].val);
            morphs[11].SetValue (clamp (rUp * spreading.val * spreadingInner.val + biasInner.val, minSpread.val, maxSpread.val) + morphValues[11].val);
        }

        private void OnDestroy () {
            for (int i = 0; i < morphs.Length; ++i) {
                if (morphs[i] != null) {
                    // Disable morph locks
                    if (morphs[i].drivenBy == pluginName) {
                        morphs[i].SetDriven (false, pluginName, true);
                    }
                    // Restore morph values
                    if (morphValues[i] != null) {
                        // morphValues have some weird internal and external memory. This hack makes sure the value is set in both.
                        morphs[i].morphValue = morphValues[i].val + .1f;
                        morphs[i].SyncJSON ();
                        morphs[i].morphValue = morphValues[i].val;
                    }
                }
            }
        }

        private void LogError (string error) {
            SuperController.LogError ($"[{pluginName}] {error}");
        }

        private JSONStorableFloat CreateFloatParam (string label, float defaultVal, float min, float max, bool right = false, JSONStorableFloat.SetFloatCallback callback = null) {
            JSONStorableFloat newFloat;
            if (callback != null) {
                newFloat = new JSONStorableFloat (label, defaultVal, callback, min, max, false, true);
            } else {
                newFloat = new JSONStorableFloat (label, defaultVal, min, max, false, true);
            }
            RegisterFloat (newFloat);
            CreateSlider (newFloat, right);
            return newFloat;
        }

        private float clamp (float val, float min, float max) {
            if (val < min) return min;
            if (val > max) return max;
            return val;
        }
    }
}